function [X, map] = readpsm(filename)
%READPSM Read image data from a PSM file.
%
%   [X, MAP] = READPSM(FILENAME) reads image data from a PSM file.
%   X is an M-by-N array. PSM images have no colormap so MAP is always empty.
%

%   Multi-byte color component values are always stored using
%   big-endian byte order.
%
% Author:      Carlo Tomasi
% Modified from readpnm by
% Author:      Peter J. Acklam
% E-mail:      pjacklam@online.no

%  READPNM Copyright 2001-2002 The MathWorks, Inc.
%  $Revision: 1.1.6.2 $  $Date: 2004/02/01 22:04:51 $

   info = impsminfo(filename);

   X     = [];
   map   = [];

   filename = info.Filename;
   format   = lower(info.Format);
   width    = info.Width;
   height   = info.Height;
   maxval   = info.MaxValue;
   encoding = info.Encoding;
   offset   = info.ImageDataOffset;

   % Number of channels (planes) in the image.
   channels = 1;

   % Open file for reading.  Multi-byte format is big-endian.
   [fid, msg] = fopen(filename, 'r', 'ieee-be');
   if fid < 0
      error('MATLAB:readpnm:fileOpen', ...
            [filename ': ' msg]);
   end

   % Seek forward to the image data.
   fseek(fid, offset, 0);

   % Get precision string for FREAD and number of bytes pr color
   % component value.
   if maxval <= 255
       precision = '*uint8';
   elseif maxval <= 65535
       precision = '*uint16';
   else
       error('MATLAB:readpsm:badMaxval', ...
           '%s: maxval of %d not supported', filename, maxval);
   end

   % Figure out how many values FREAD must read.

   % numVals - number of values required to store all scanlines
   numVals = height*width*channels;

   % Read the data.
   [X, count] = fread(fid, numVals, precision);

   % The next read should not return any data.  If it does, then there is
   % trailing garbage at the end of the file.  This may indicate a
   % corrupt image (e.g., that each LF has been converted to CR+LF).
   if length(fread(fid, 1, 'char'));
       warning('MATLAB:readpsm:extraData', ...
           [filename ': file contains extra data.']);
   end

   % Close the file.
   fclose(fid);

   if count < numVals
       warning('MATLAB:readpsm:unexpectedEOF', ...
           [filename ': file ended while reading image data.']);
       % Fill in the missing values with zeros.
       X(numVals) = 0;
   end

   % Reshape arrays to correct size.
   X = reshape(X, [channels width height]);
   X = permute(X, [3 2 1]);

   % With MATLAB images, the the minimum pixel value is always zero, the
   % maximum pixel value is 255 for uint8 arrays, 65535 for uint16 arrays
   % and 1 for double arrays.  In PSM images, the maximum pixel
   % value might be any positive integer, so we might need to scale the
   % pixel values.  If MAXVAL = 2^N-1, for some integer N, we use bit
   % shuffling to avoid temporary conversion to double.  Otherwise we
   % convert to double and perform the necessary arithmetic operations.

   if maxval <= 255
      % Scale from 0..MAXVAL to 0..255 and store as uint8.
      switch maxval
         case 1         % = 2^1-1
            X = bitor(bitshift(X, 1), X);                   % 1 -> 3
            X = bitor(bitshift(X, 2), X);                   % 3 -> 15
            X = bitor(bitshift(X, 4), X);                   % 15 -> 255
         case 3         % = 2^2-1
            X = bitor(bitshift(X, 2), X);                   % 3 -> 15
            X = bitor(bitshift(X, 4), X);                   % 15 -> 255
         case 7         % = 2^3-1
            X = bitor(bitshift(X, 3), X);                   % 7 -> 63
            X = bitor(bitshift(X, 2), bitshift(X, -4));     % 63 -> 255
         case 15        % = 2^4-1
            X = bitor(bitshift(X, 4), X);                   % 15 -> 255
         case 31        % = 2^5-1
            X = bitor(bitshift(X, 3), bitshift(X, -2));     % 31 -> 255
         case 63        % = 2^6-1
            X = bitor(bitshift(X, 2), bitshift(X, -4));     % 63 -> 255
         case 127       % = 2^7-1
            X = bitor(bitshift(X, 1), bitshift(X, -6));     % 127 -> 255
         case 255       % = 2^8-1
            % nothing to do
         otherwise
            X = uint8(round(255/maxval*double(X)));
      end
   elseif maxval <= 65535
      % Scale from 0..MAXVAL to 0..65535 and store as uint16.
      switch maxval
         case 511       % = 2^9-1
            X = bitor(bitshift(X, 7), bitshift(X, -2));     % 511 -> 65535
         case 1023      % = 2^10-1
            X = bitor(bitshift(X, 6), bitshift(X, -4));     % 1023 -> 65535
         case 2047      % = 2^11-1
            X = bitor(bitshift(X, 5), bitshift(X, -6));     % 2047 -> 65535
         case 4095      % = 2^12-1
            X = bitor(bitshift(X, 4), bitshift(X, -8));     % 4095 -> 65535
         case 8191      % = 2^13-1
            X = bitor(bitshift(X, 3), bitshift(X, -10));    % 8191 -> 65535
         case 16383     % = 2^14-1
            X = bitor(bitshift(X, 2), bitshift(X, -12));    % 16383 -> 65535
         case 32767     % = 2^15-1
            X = bitor(bitshift(X, 1), bitshift(X, -14));    % 32767 -> 65535
         case 65535     % = 2^16-1
            % nothing to do
         otherwise
            X = uint16(round(65535/maxval*double(X)));
      end
   else
      % Scale from 0..MAXVAL to 0..1 and store as double.
      X = double(X)/maxval;
   end
